
from generallibrary import Markdown, current_datetime_formatted


class _PackagerMarkdown:
    """ Contains methods to generate readme sections from arguments.
        Todo: Inherit future crawler class for pypi and github. """
    def get_badges_dict(self):
        """ Get badges as a dict.

            :param generalpackager.Packager self: """
        return {
            "UnitTests": f"[![workflow Actions Status](https://github.com/ManderaGeneral/{self.name}/workflows/workflow/badge.svg)](https://github.com/ManderaGeneral/{self.name}/actions)",
            "Commit": f"![GitHub last commit](https://img.shields.io/github/last-commit/ManderaGeneral/{self.name})",
            "Release": f"[![PyPI version shields.io](https://img.shields.io/pypi/v/{self.name}.svg)](https://pypi.org/project/{self.name}/)",
            "Python": f"[![PyPI pyversions](https://img.shields.io/pypi/pyversions/{self.name}.svg)](https://pypi.python.org/pypi/{self.name}/)",
            "Operating System": f"[![Generic badge](https://img.shields.io/badge/platforms-{'%20%7C%20'.join(self.os)}-blue.svg)](https://shields.io/)",
        }

    def get_description_markdown(self):
        """ Get information table.

            :param generalpackager.Packager self: """
        part_of = f"This package and {len(self.packagers_dict) - 1} other make up {Markdown.link(text='ManderaGeneral', url='https://github.com/Mandera')}."
        header = f"{self.name} {self.localrepo.version}"
        return Markdown(self.localrepo.description, "\n", part_of, header=header)

    def get_information_markdown(self, *packagers):
        """ Get information table.

            :param generalpackager.Packager self: """
        if not packagers:
            packagers = (self, )

        markdown = Markdown(header="Information")

        python_url = "https://www.python.org/downloads/release/python-"

        list_of_dicts = []
        for packager in packagers:
            list_of_dicts.append({
                "Package": Markdown.link(text=packager.name, url=packager.github.url),
                "Version": Markdown.link(text=packager.localrepo.version, url=packager.pypi.url),
                "Latest Release": packager.pypi.get_datetime(),
                "Python": ", ".join([Markdown.link(text=ver, url=f"{python_url}{str(ver).replace('.', '')}0/") for ver in packager.python]),
                "Platform": ", ".join(map(str.capitalize, packager.os)),
            })
        markdown.add_table_lines(*list_of_dicts, sort_by="Package")
        return markdown

    def get_installation_markdown(self):
        """ Get install markdown.

            :param generalpackager.Packager self: """
        markdown = Markdown(header="Installation")

        dependencies_required = self.localrepo.install_requires.copy()
        dependencies_optional = list(set().union(*self.localrepo.extras_require.values()))
        dependencies_optional.sort()

        options = {self.name: dependencies_required}
        options.update({f"{self.name}[{key}]": value + dependencies_required for key, value in self.localrepo.extras_require.items()})

        list_of_dicts = []

        for command, packages in options.items():
            row = {"Command": f"`pip install {command}`"}
            for dependency in dependencies_required + dependencies_optional:
                row[Markdown.link(dependency, url=f"https://pypi.org/project/{dependency}", href=True)] = "Yes" if dependency in packages else "No"
            list_of_dicts.append(row)

        markdown.add_table_lines(*list_of_dicts)

        return markdown

    def configure_contents_markdown(self, markdown):
        """ Configure table of contents lines from markdown.

            :param generalpackager.Packager self:
            :param markdown: """
        parent_markdown = markdown.get_parent(-1)
        markdown.add_pre_lines(parent_markdown.view(custom_repr=lambda md: md.link(md.header, href=True), print_out=False))
        return markdown

    def _attr_repr(self, objInfo):
        """ Return a nice representation of each attribute made by this module, in this case a link to code definition.

            :param generalpackager.Packager self:
            :param generallibrary.ObjInfo objInfo: """
        text = objInfo.nice_repr()
        commit_sha = self.commit_sha
        file_path = f'{objInfo.module().__name__.replace(".", "/")}{"/__init__" if objInfo.is_module() else ""}.py'
        line = objInfo.get_definition_line()

        return self.github_link(text=text, suffix=f"blob/{commit_sha}/{file_path}#L{line}")

    def github_link(self, text, suffix):
        """ :param generalpackager.Packager self:
            :param text:
            :param suffix: """
        url = f"{self.github.url}/{suffix}"
        # self.assert_url_up(url=url)  # Wont work for private repos or new files, would have to check after the fact.
        return Markdown.link(text=text, url=url, href=True)

    def get_attributes_markdown(self):
        """ Get a recursive view of attributes markdown.

            :param generalpackager.Packager self: """
        view_str = self.localmodule.objInfo.view(custom_repr=self._attr_repr, print_out=False)
        return Markdown(header="Attributes").add_pre_lines(view_str)

    def get_footnote_markdown(self, commit=True):
        """ Get a markdown for footnote containing date, time and commit link.

            :param generalpackager.Packager self:
            :param commit: """
        line = f"Generated {current_datetime_formatted()}"
        if commit:
            line += f" for commit {self.github_link(text=self.commit_sha, suffix=f'commit/{self.commit_sha}')}."

        return Markdown(line).wrap_with_tags("sup")





from generallibrary import NetworkDiagram

from generalpackager import Packager
from generalpackager.api.local_repo import LocalRepo


# class PackagerNetwork(NetworkDiagram):
#     def __init__(self, packager):
#         self.packager = packager


# class PackagerGrp(list):
#     """ Handles a collection of packages.
#         Todo: Maybe move PackagerGrp to Packager now that it inherits NetworkDiagram? """
#     def __init__(self, *packagers, repos_path=None):
#         super().__init__(packagers)
#
#         self.repos_path = LocalRepo.get_repos_path(path=repos_path)
#
#         self.load_general_packages()
#
#     def load_general_packages(self):
#         """ Load my general packages. """
#         self.add_packages_from_names(*Packager.get_users_package_names())
#
#     def add_packages_from_names(self, *names):
#         """ Add a Package. """
#         self.extend([Packager(name=name, repos_path=self.repos_path) for name in names])
#
#     def clone(self):
#         """ Clone all packages to repos_path. """
#         for packager in self:
#             packager.clone_repo()
#
#     def install(self):
#         """ Install all packages. """
#         for packager in self:
#             packager.localrepo.pip_install()
#
#     def get_bumped(self):
#         """ Get a list of bumped packagers, meaning PyPI version and LocalRepo version mismatch. """
#         # return [(packager.localrepo.version, packager.pypi.get_version()) for packager in self.packagers if packager.is_bumped()]
#         return [packager for packager in self if packager.is_bumped()]








class _PackagerRelations:
    packagers_dict = {}

    def add(self):
        """ Add this packager to packagers dict.

            :param generalpackager.Packager self: """
        if self.name in self.packagers_dict:
            raise AttributeError(f"{self.name} packager already exists")

        self.packagers_dict[self.name] = self

    def remove(self):
        """ Remove this packager from packagers dict.

            :param generalpackager.Packager self: """
        self.packagers_dict.pop(self.name)

    def update_links(self):
        """ Update links of all created Packagers.

            :param generalpackager.Packager self: """
        for packager in self.packagers_dict.values():
            for dependency_name in packager.localrepo.install_requires:
                dependency_packager = self.packagers_dict.get(dependency_name)
                if dependency_packager is not None:
                    dependency_packager.link(target=packager)

    def get_packager_with_name(self, name):
        """ Return connected Packager or None.

            :param generalpackager.Packager self:
            :param name: """
        if self.name == name:
            return self

        packager = self.packagers_dict.get(name)
        if packager is None and self.is_creatable(name=name):
            packager = type(self)(name=name, repos_path=self.repos_path)

        return packager

    def load_general_packagers(self):
        """ Load packagers with names.

            :param generalpackager.Packager self: """
        if not self.packagers_dict:
            for name in self.get_users_package_names():
                packager = self.get_packager_with_name(name=name)

                if packager and packager.localrepo.enabled:
                    packager.add()
                    self.update_links()
        return self.packagers_dict

    def get_dependencies(self):
        """ Get set of loaded Packagers that this Packager requires.

            :param generalpackager.Packager self: """
        return self.get_nodes(incoming=True, outgoing=False)

    def get_dependents(self):
        """ Get set of loaded Packagers that requires this Packager.

            :param generalpackager.Packager self: """
        return self.get_nodes(incoming=False, outgoing=True)

    def get_ordered_packagers(self):
        """ Get a list of ordered names from the dependency chain.

            :param generalpackager.Packager self: """
        return [packager for packager_set in self.get_ordered() for packager in sorted(packager_set, key=lambda x: x.name)]

    @classmethod
    def get_users_package_names(cls, pypi_user=None, github_user=None):
        """ Return a set of user's packages with intersecting PyPI and GitHub.

            :param generalpackager.Packager cls:
            :param pypi_user:
            :param github_user: """
        return cls.PyPI.get_users_packages(user=pypi_user).intersection(cls.GitHub.get_users_packages(user=github_user))

    def general_bumped_set(self):
        """ Return a set of general packagers that have been bumped.

            :param generalpackager.Packager self: """
        self.load_general_packagers()
        return {packager for packager in self.packagers_dict.values() if packager.is_bumped()}

    def general_changed_dict(self, aesthetic=False):
        """ Return a dict of general packagers with changed files comparing to remote.

            :param generalpackager.Packager self:
            :param aesthetic: """
        self.load_general_packagers()
        return {packager: files for packager in self.packagers_dict.values() if (files := packager.compare_local_to_remote(aesthetic=aesthetic))}





from generallibrary import Ver, deco_cache
from generalfile import Path

import requests
import re
import shutil


def download(url, path):
    """ Todo: Move download to it's own package.
        Todo: Binary extension for generalfile. """
    data = requests.get(url)
    path = Path(path)

    with path.lock():
        path.get_parent().create_folder()
        with open(str(path), "wb") as file:
            file.write(data.content)


def unpack(path, target):
    """ Todo: Add pack and unpack to generalfile. """
    shutil.unpack_archive(str(path), str(target))


class PyPI:
    """ Tools to interface pypi.org """
    def __init__(self, name):
        self.name = name

        self.url = self.get_url(name=self.name)

    @classmethod
    def is_creatable(cls, name):
        """ Return whether this API can be created. """
        return requests.get(url=cls.get_url(name=name)).status_code == 200

    @staticmethod
    def get_url(name):
        """ Get static URL from owner and name. """
        return f"https://pypi.org/project/{name}/"

    @deco_cache()
    def get_tarball_url(self, name=None, version=None):
        """ Get URL to download tarball. """
        if name is None:
            name = self.name
        if version is None:
            version = self.get_version(name=name)
        return f"https://pypi.io/packages/source/{name[0]}/{name}/{name}-{version}.tar.gz"

    def download_and_unpack_tarball(self, target_folder, name=None, version=None):
        """ Download tar ball, extract it, remove tar ball. """
        temp = Path.get_cache_dir() / "Python/temp.tar.gz"
        download(self.get_tarball_url(name=name, version=version), path=temp)
        unpack(path=temp, target=target_folder)
        temp.delete(error=False)
        return target_folder

    @staticmethod
    @deco_cache()
    def get_users_packages(user=None):
        """ Get a set of a user's packages' names on PyPI. """
        if user is None:
            user = "Mandera"
        return set(re.findall("/project/(.*)/", requests.get(f"https://pypi.org/user/{user}/").text))

    @deco_cache()
    def get_version(self, name=None):
        """ Get version of latest publish on PyPI. """
        if name is None:
            name = self.name
        return Ver(re.findall(f"{name} ([.0-9]+)\n", requests.get(f"https://pypi.org/project/{name}/").text)[0])

    @deco_cache()
    def get_datetime(self, name=None):
        """ Get datetime of latest release.
            Todo: Proper date fetch. """
        if name is None:
            name = self.name
        requests.get(f"https://pypi.org/project/{name}/")
        result = re.findall('Generated (.+) for commit', requests.get(f"https://pypi.org/project/{name}/").text)
        return result[0] if result else "-"

    def reserve_name(self):
        pass





























""" Methods specific for my general packages.
    Isolatable methods are put inside APIs. """

from generallibrary import initBases, NetworkDiagram
from generalpackager import LocalRepo, LocalModule, GitHub, PyPI

from generalpackager.packager_files import _PackagerFiles
from generalpackager.packager_github import _PackagerGitHub
from generalpackager.packager_markdown import _PackagerMarkdown
from generalpackager.packager_metadata import _PackagerMetadata
from generalpackager.packager_pypi import _PackagerPypi
from generalpackager.packager_workflow import _PackagerWorkflow
from generalpackager.packager_relations import _PackagerRelations


@initBases
class Packager(NetworkDiagram, _PackagerMarkdown, _PackagerGitHub, _PackagerFiles, _PackagerMetadata, _PackagerPypi, _PackagerWorkflow, _PackagerRelations):
    """ Uses APIs to manage 'general' package.
        Contains methods that require more than one API as well as methods specific for ManderaGeneral.
        Todo: Allow github, pypi or local repo not to exist in any combination.
        Todo: Replace badges with generated hardcode.
        Todo: Find all imports to generate install_requires.
        Todo: Create links for Todos. """

    LocalRepo, LocalModule, GitHub, PyPI = LocalRepo, LocalModule, GitHub, PyPI

    author = 'Rickard "Mandera" Abraham'
    email = "rickard.abraham@gmail.com"
    license = "mit"
    python = "3.8", "3.9"  # Only supports basic definition with tuple of major.minor.
    os = "windows", "ubuntu"  # , "macos"

    git_exclude_lines = ".idea", "build", "dist", "*.egg-info", "__pycache__", ".git"

    def __init__(self, name, repos_path=None, owner=None):
        if owner is None:
            owner = "ManderaGeneral"

        self.name = name
        self.repos_path = LocalRepo.get_repos_path(path=repos_path)
        self.owner = owner

        self.path = self.repos_path / self.name

        self._localrepo = None
        self._github = None
        self._localmodule = None
        self._pypi = None

    def is_creatable(self, name):
        """ Simple placeholder check. """
        # return LocalRepo.is_creatable(path=self.repos_path / name)
        return GitHub.is_creatable(name=name, owner="ManderaGeneral")

    @property
    def localrepo(self):
        """ Generate, protect and cache. """
        if not self._localrepo:
            if not LocalRepo.is_creatable(path=self.path):
                if self.github:
                    self.clone_repo(url=self.github.url, path=self.path)
                else:
                    return None
            self._localrepo = LocalRepo(path=self.path, git_exclude_lines=self.git_exclude_lines)

        return self._localrepo

    @property
    def github(self):
        """ Generate, protect and cache. """
        if not self._github:
            self._github = GitHub(name=self.name, owner=self.owner)

        return self._github

    @property
    def localmodule(self):
        """ Generate, protect and cache. """
        if not self._localmodule:
            self._localmodule = LocalModule(name=self.name)

        return self._localmodule

    @property
    def pypi(self):
        """ Generate, protect and cache. """
        if not self._pypi:
            self._pypi = PyPI(name=self.name)
        return self._pypi

    def generate_localfiles(self, aesthetic=True):
        """ Generate all local files. """
        for generate in self.files:
            if aesthetic or not generate.aesthetic:
                generate.generate()

    def __repr__(self):
        return f"<Packager: {self.name}>"






















































from generalpackager import PACKAGER_GITHUB_API

import requests
import json
import re


class GitHub:
    """ Tools to interface a GitHub Repository. """
    def __init__(self, name, owner=None):
        if owner is None:
            owner = "ManderaGeneral"

        self.name = name
        self.owner = owner

        self.url = self.get_url(name=self.name, owner=self.owner)

        # if not self.is_url_functioning():
        #     raise AssertionError(f"Url for {self.name} is not functioning.")

    @classmethod
    def is_creatable(cls, name, owner):
        """ Return whether this API can be created. """
        return requests.get(url=cls.get_url(name=name, owner=owner)).status_code == 200

    @staticmethod
    def get_url(name, owner):
        """ Get static URL from owner and name. """
        return f"https://github.com/{owner}/{name}"

    def is_url_functioning(self, url=None):
        """ Checks name, owner and token all in one. """
        return self._request(url=url).status_code == 200

    def api_url(self, endpoint=None):
        """ Get URL from owner, name and enpoint. """
        return "/".join(("https://api.github.com", "repos", self.owner, self.name) + ((endpoint, ) if endpoint else ()))

    def get_website(self):
        """ Get website specified in repository details.

            :rtype: list[str] """
        return self._request(method="get").json()["homepage"]

    def set_website(self, website):
        """ Set a website for the GitHub repository. """
        return self._request(method="patch", name=self.name, homepage=website)


    def get_topics(self):
        """ Get a list of topics in the GitHub repository.

            :rtype: list[str] """
        return self._request(method="get", endpoint="topics").json()["names"]

    def set_topics(self, *topics):
        """ Set topics for the GitHub repository.

            :param str topics: """
        return self._request(method="put", endpoint="topics", names=topics)


    def get_description(self):
        """ Get a string of description in the GitHub repository.

            :rtype: list[str] """
        return self._request(method="get").json()["description"]

    def set_description(self, description):
        """ Set a description for the GitHub repository. """
        return self._request(method="patch", name=self.name, description=description)

    def _request(self, method="get", url=None, endpoint=None, **data):
        """ :rtype: requests.Response """
        method = getattr(requests, method.lower())

        kwargs = {
            "headers": {"Accept": "application/vnd.github.mercy-preview+json"},
            "auth": (self.owner, PACKAGER_GITHUB_API.value),
        }
        if data:
            kwargs["data"] = json.dumps(data)

        if url is None:
            url = self.api_url(endpoint=endpoint)
        return method(url=url, **kwargs)

    @staticmethod
    def get_users_packages(user=None):
        """ Get a set of a user's packages' names on GitHub. """
        if user is None:
            user = "ManderaGeneral"
        return set(re.findall(f'"/{user}/([a-z]*)"', requests.get(f"https://github.com/{user}?tab=repositories").text))



















from generalfile import Path
from generalpackager import GIT_PASSWORD
from generallibrary import Ver

from setuptools import find_namespace_packages
import re
from git import Repo
import subprocess
import sys


class LocalRepo:
    """ Tools to help Path interface a Local Python Repository. """
    enabled = True
    name = ...
    version = "0.0.1"  # type: Ver
    description = "Missing description."
    install_requires = []
    extras_require = {}
    topics = []
    manifest = []

    metadata_keys = [key for key, value in locals().items() if not key.startswith("_")]

    def __init__(self, path, git_exclude_lines):
        assert self.path_is_repo(path=path)

        self.path = Path(path).absolute()
        self.git_exclude_lines = git_exclude_lines

        self.has_metadata = self.get_metadata_path().exists(quick=True)
        if self.has_metadata:
            for key, value in self.get_metadata_path().read().items():
                setattr(self, f"_{key}", value)

        if self.extras_require:
            self.extras_require["full"] = list(set().union(*self.extras_require.values()))
            self.extras_require["full"].sort()

        self.version = Ver(self.version)

        if self.name is Ellipsis:
            self.name = self.path.name()

    @staticmethod
    def get_repos_path(path):
        """ Try to return repos path by iterating parents if None. """
        if path is None:
            repos_path = Path.get_working_dir()
            while not LocalRepo.get_local_repos(repos_path):
                repos_path = repos_path.get_parent()
                if repos_path is None:
                    raise AttributeError(f"Couldn't find repos path.")
            return repos_path
        else:
            return Path(path).absolute()

    @classmethod
    def is_creatable(cls, path):
        """ Return whether this API can be created. """
        return cls.path_is_repo(path=path)

    def metadata_setter(self, key, value):
        """ Set a metadata's key both in instance and json file. """
        if self.has_metadata and value != getattr(self, f"_{key}", ...):
            metadata = self.get_metadata_path().read()
            metadata[key] = str(value)
            self.get_metadata_path().write(metadata, overwrite=True, indent=4)

        setattr(self, f"_{key}", value)

    def get_readme_path(self):
        """ Get a Path instance pointing to README.md, regardless if it exists. """
        return self.path / "README.md"

    def get_metadata_path(self):
        """ Get a Path instance pointing to metadata.json, regardless if it exists. """
        return self.path / "metadata.json"

    def get_git_exclude_path(self):
        """ Get a Path instance pointing to .git/info/exclude, regardless if it exists. """
        return self.path / ".git/info/exclude"

    def get_setup_path(self):
        """ Get a Path instance pointing to setup.py, regardless if it exists. """
        return self.path / "setup.py"

    def get_manifest_path(self):
        """ Get a Path instance pointing to MANIFEST.in, regardless if it exists. """
        return self.path / "MANIFEST.in"

    def get_license_path(self):
        """ Get a Path instance pointing to LICENSE, regardless if it exists. """
        return self.path / "LICENSE"

    def get_workflow_path(self):
        """ Get a Path instance pointing to workflow.yml, regardless if it exists. """
        return self.path / ".github/workflows/workflow.yml"

    def get_test_path(self):
        """ Get a Path instance pointing to workflow.yml, regardless if it exists. """
        return self.path / f"{self.name}/test"

    def get_package_paths(self):
        """ Get a list of Paths pointing to each folder containing a Python file in this local repo, aka `namespace package`. """
        return [self.path / pkg.replace(".", "/") for pkg in find_namespace_packages(where=str(self.path))]

    @classmethod
    def get_local_repos(cls, folder_path):
        """ Return a list of local repos in given folder. """
        folder_path = Path(folder_path)
        if not folder_path.exists():
            return []
        return [path for path in folder_path.get_paths_in_folder() if cls.path_is_repo(path)]

    @classmethod
    def path_is_repo(cls, path):
        """ Return whether this path is a local repo. """
        path = Path(path)
        if path.is_file() or not path.exists(quick=True):
            return False
        return ".git" in map(Path.name, path.get_paths_in_folder())

    def get_todos(self):
        """ Get a list of dicts containing cleaned up todos.

            :rtype: dict[list[str]] """
        todos = []
        for path in self.path.get_paths_recursive():
            if path.match(*self.git_exclude_lines, "shelved.patch", "readme.md"):
                continue
            try:
                text = path.text.read()
            except:
                continue

            for todo in re.findall("todo+: (.+)", text, re.I):
                todos.append({
                    "Module": path.name(),
                    "Message": re.sub('[" ]*$', "", todo),
                })
        return todos

    def commit_and_push(self, message=None, tag=False, owner=None):
        """ Commit and push this local repo to GitHub.
            Return short sha1 of pushed commit. """
        if message is None:
            message = "Automatic commit."
        if owner is None:
            owner = "ManderaGeneral"

        repo = Repo(str(self.path))

        repo.git.add(A=True)
        repo.index.commit(message=str(message))
        remote = repo.remote()
        remote.set_url(f"https://Mandera:{GIT_PASSWORD}@github.com/{owner}/{self.name}.git")

        if tag:
            tag_ref = repo.create_tag(f"v{self.version}", force=True)
            remote.push(refspec=tag_ref)

        return remote.push()[0].summary.split("..")[1].rstrip()

    def get_changed_files(self):
        """ Get a list of changed files compared to remote. """
        repo = Repo(str(self.path))
        return re.findall("diff --git a/(.*) " + "b/", repo.git.diff())

    def bump_version(self):
        """ Bump micro version in metadata.json. """
        self.version = self.version.bump()

    def pip_install(self):
        """ Install this repository with pip, WITHOUT -e flag.
            Subprocess messed up -e flag compared to doing it in terminal, so use the normal one."""
        subprocess.check_call([sys.executable, "-m", "pip", "install", str(self.path)])

    def unittest(self):
        """ Run unittests for this repository. """
        subprocess.check_call([sys.executable, "-m", "unittest", "discover", str(self.get_test_path())])

    def create_sdist(self):
        """ Create source distribution. """
        with self.path.as_working_dir():
            subprocess.check_call([sys.executable, "setup.py", "sdist", "bdist_wheel"])

    def upload(self):
        """ Upload local repo to PyPI. """
        self.create_sdist()
        with self.path.as_working_dir():
            subprocess.check_call([sys.executable, "-m", "twine", "upload", "dist/*"])

for key in LocalRepo.metadata_keys:
    value = getattr(LocalRepo, key)
    setattr(LocalRepo, key, property(
        fget=lambda self, key=key, value=value: getattr(self, f"_{key}", value),
        fset=lambda self, value, key=key: LocalRepo.metadata_setter(self, key, value),
    ))






































from generallibrary import CodeLine, current_datetime, Markdown
from generalfile import Path


class GenerateFile:
    """ Handle generation of files. """
    def __init__(self, path, text_func, packager, aesthetic):
        self.text_func = text_func
        self.packager = packager
        self.aesthetic = aesthetic

        self.relative_path = path.relative(base=packager.path)
        self.path = packager.path / self.relative_path

    def generate(self):
        """ Generate actual file. """
        self.path.text.write(f"{self.text_func()}\n", overwrite=True)


class _PackagerFiles:
    """ Generates setup, license and gitexclude.
        Only changed non-aesthetic files can trigger a version bump and publish. """
    extra_aesthetic = "randomtesting.py",  # "licenses"
    extra_non_aesthetic = tuple()

    def __init_post__(self):
        """ Todo: Watermark generated files to prevent mistake of thinking you can modify them directly.

            :param generalpackager.Packager self: """
        self.file_setup =           GenerateFile(self.localrepo.get_setup_path(), self.generate_setup, self, aesthetic=False)
        self.file_manifest =        GenerateFile(self.localrepo.get_manifest_path(), self.generate_manifest, self, aesthetic=False)

        self.file_git_exclude =     GenerateFile(self.localrepo.get_git_exclude_path(), self.generate_git_exclude, self, aesthetic=True)
        self.file_license =         GenerateFile(self.localrepo.get_license_path(), self.generate_license, self, aesthetic=True)
        self.file_workflow =        GenerateFile(self.localrepo.get_workflow_path(), self.generate_workflow, self, aesthetic=True)
        self.file_readme =          GenerateFile(self.localrepo.get_readme_path(), self.generate_readme, self, aesthetic=True)

        self.files = [getattr(self, key) for key in dir(self) if key.startswith("file_")]
        self.files_by_relative_path = {file.relative_path: file for file in self.files}

        self.file_personal_readme = GenerateFile(self.localrepo.get_readme_path(), self.generate_personal_readme, self, aesthetic=True)

    def relative_path_is_aesthetic(self, relative_path):
        """ Relative to package path. None if not defined as a GenerateFile instance.

            :param generalpackager.Packager self:
            :param Path relative_path: """
        aesthetic_attr = getattr(self.files_by_relative_path.get(relative_path, None), "aesthetic", None)
        if aesthetic_attr is None:
            if relative_path.match(*self.extra_aesthetic):
                return True
            elif relative_path.match(*self.extra_non_aesthetic):
                return False
        return aesthetic_attr

    def filter_relative_filenames(self, *filenames, aesthetic, non_defined=True):
        """ If aesthetic is None then it doesn't filter any.
            True will return only aesthetic.
            False will return only non-aesthetic.

            :param generalpackager.Packager self:
            :param bool or None aesthetic:
            :param bool non_defined: """
        result = []
        for filename in filenames:
            is_aesthetic = self.relative_path_is_aesthetic(filename)
            if non_defined is False and is_aesthetic is None:
                continue
            if aesthetic is True and is_aesthetic is False:
                continue
            if aesthetic is False and is_aesthetic is True:
                continue
            result.append(filename)
        return result

    def compare_local_to_remote(self, aesthetic=None):
        """ Get a list of changed files compared to remote with optional aesthetic files.

            :param generalpackager.Packager self:
            :param aesthetic: """
        return self.filter_relative_filenames(*self.localrepo.get_changed_files(), aesthetic=aesthetic)

    def generate_setup(self):
        """ Generate setup.py.

            :param generalpackager.Packager self: """
        readme_path = self.localrepo.get_readme_path().relative(self.localrepo.get_setup_path().get_parent())
        last_version_split = self.python[-1].split(".")
        last_version_bumped_micro = f"{last_version_split[0]}.{int(last_version_split[1]) + 1}"
        setup_kwargs = {
            "name": f'"{self.localrepo.name}"',
            "author": f"'{self.author}'",
            "author_email": f'"{self.email}"',
            "version": f'"{self.localrepo.version}"',
            "description": f'"{self.localrepo.description}"',
            "long_description": f"(Path(__file__).parent / '{readme_path}').read_text(encoding='utf-8')",
            "long_description_content_type": '"text/markdown"',
            "install_requires": self.localrepo.install_requires,
            "url": f'"{self.github.url}"',
            "license": f'"{self.license}"',
            "python_requires": f'">={self.python[0]}, <{last_version_bumped_micro}"',
            "packages": 'find_namespace_packages(exclude=("build*", "dist*"))',
            "extras_require": self.localrepo.extras_require,
            "classifiers": self.get_classifiers(),
            # "include_package_data": True,
        }

        top = CodeLine()
        top.add(CodeLine("from setuptools import setup, find_namespace_packages", space_before=1))
        top.add(CodeLine("from pathlib import Path", space_after=1))

        setup = top.add(CodeLine("setup("))
        for key, value in setup_kwargs.items():
            if isinstance(value, list) and value:
                list_ = setup.add(CodeLine(f"{key}=["))
                for item in value:
                    list_.add(CodeLine(f"'{item}',"))
                setup.add(CodeLine("],"))
            elif isinstance(value, dict) and value:
                dict_ = setup.add(CodeLine(f"{key}={{"))
                for k, v in value.items():
                    dict_.add(CodeLine(f"'{k}': {v},"))
                setup.add(CodeLine("},"))
            else:
                setup.add(CodeLine(f"{key}={value},"))

        top.add(CodeLine(")"))

        return top.text()

    def generate_manifest(self):
        """ Generate manifest file.

            :param generalpackager.Packager self: """
        default_manifest = [
            self.localrepo.get_metadata_path().relative(self.path),
        ]
        return "\n".join([f"include {path}" for path in self.localrepo.manifest + default_manifest])

    def generate_git_exclude(self):
        """ Generate git exclude file.

            :param generalpackager.Packager self: """
        return "\n".join(self.git_exclude_lines)

    def generate_license(self):
        """ Generate LICENSE by using Packager.license.

            :param generalpackager.Packager self: """
        text = Path(self.repos_path / f"generalpackager/generalpackager/licenses/{self.license}").text.read()
        assert "$" in text
        text = text.replace("$year", str(current_datetime().year))
        text = text.replace("$author", self.author)
        assert "$" not in text

        return text

    def generate_workflow(self):
        """ Generate workflow.yml.

            :param generalpackager.Packager self: """
        workflow = CodeLine()
        workflow.indent_str = " " * 2

        workflow.add("name: workflow")
        workflow.add(self.get_triggers())

        jobs = workflow.add("jobs:")
        jobs.add(self.get_unittest_job())
        jobs.add(self.get_sync_job())

        return workflow.text()

    def generate_readme(self):
        """ Generate readme markdown and overwrite README.md in local repo.

            :param generalpackager.Packager self: """
        # Description
        markdown = self.get_description_markdown()

        # Information
        self.get_information_markdown().set_parent(parent=markdown)

        # Table of contents
        contents = Markdown(header="Contents", parent=markdown)

        # Installation
        self.get_installation_markdown().set_parent(parent=markdown)

        # Attributes
        self.get_attributes_markdown().set_parent(parent=markdown)

        # Todos
        todos = self.localrepo.get_todos()
        if todos:
            Markdown(header=f"Todo ({len(todos)}x)", parent=markdown).add_table_lines(*todos)

        # Table of contents - Configuration
        self.configure_contents_markdown(markdown=contents)

        # Footnote
        self.get_footnote_markdown().set_parent(parent=markdown)

        return markdown

    def generate_personal_readme(self):
        """ Generate personal readme markdown and overwrite README.md in local repo.

            :param generalpackager.Packager self: """
        # Description
        markdown = Markdown(header="General").add_list_lines(
            "A collection of connected packages.",
            "Violently updated with little regard for backwards compatability.",
            "Automatic workflows to unittest, sync and publish.",
        )

        # Package information
        self.get_information_markdown(*self.load_general_packagers().values()).set_parent(parent=markdown)

        # Footnote
        self.get_footnote_markdown(commit=False).set_parent(parent=markdown)

        return markdown

































from git import Repo


class _PackagerGitHub:
    def __init__(self):
        self.commit_sha = "master"

    """ Sync metadata. """
    def sync_github_metadata(self):
        """ Sync GitHub with local metadata.

            :param generalpackager.Packager self: """
        self.github.set_website(self.pypi.url)
        self.github.set_description(self.localrepo.description)
        self.github.set_topics(*self.get_topics())

    def clone_repo(self, url=None, path=None):
        """ Clone a GitHub repo into a path to produce a LocalRepo.

            :param generalpackager.Packager self:
            :param url:
            :param path: """
        if url is None:
            url = self.github.url
        if path is None:
            path = self.path
        Repo.clone_from(url=url, to_path=path)

    def commit_push_store_sha(self, message, tag=False):
        """ Use LocalRepos method commit_and_push but also store short sha1.

            :param generalpackager.Packager self:
            :param message:
            :param tag: """
        self.commit_sha = self.localrepo.commit_and_push(message=message, tag=tag, owner=self.owner)

















from generalfile import Path


class _PackagerPypi:
    def compare_local_to_pypi(self, aesthetic=None):
        """ :param generalpackager.Packager self:
            :param aesthetic: """
        unpack_target = Path.get_cache_dir() / "Python"
        package_target = unpack_target / f"{self.name}-{self.pypi.get_version()}"
        filt = lambda path: not path.match(*self.git_exclude_lines)

        self.pypi.download_and_unpack_tarball(target_folder=unpack_target)
        differing_files = self.path.get_differing_files(target=package_target, filt=filt)
        return self.filter_relative_filenames(*differing_files, aesthetic=aesthetic)







import pkg_resources
from importlib import import_module

from generallibrary import ObjInfo


class LocalModule:
    """ Tools to interface a Local Python Module. """
    def __init__(self, name):
        self.name = name

        self.module = import_module(name=self.name)

        self.objInfo = ObjInfo(self.module)
        assert self.objInfo.is_module()
        self.objInfo.filters = [self._filter]
        self.objInfo.get_attrs(depth=-1)

    @classmethod
    def is_creatable(cls, name):
        """ Return whether this API can be created. """
        try:
            import_module(name=name)
        except ModuleNotFoundError:
            return False
        return True

    def _filter(self, objInfo):
        """ :param ObjInfo objInfo: """
        is_part_of_module = getattr(objInfo.module(), "__name__", "").startswith(self.module.__name__)
        return objInfo.public() and (objInfo.is_class() or objInfo.is_method()) and is_part_of_module

    def get_env_vars(self):
        """ Get a list of EnvVar instances avialable directly in module.

            :rtype: list[generallibrary.EnvVar] """
        objInfo = ObjInfo(self.module)
        objInfo.filters = [lambda objInfo: type(objInfo.obj).__name__ == "EnvVar"]
        objInfo.get_attrs()
        return [objInfo.obj for objInfo in objInfo.get_children()]

    @staticmethod
    def get_all_packages():
        """ Get a list of all available packages. """
        return [pkg.project_name for pkg in pkg_resources.working_set]

    def get_dependencies(self, name=None):
        """ Get a list of dependencies' names this module has. """
        if name is None:
            name = self.module.__name__
        return list(map(str, pkg_resources.working_set.by_key[name.lower()].requires()))

    def get_dependants(self, name=None):
        """ Get a list of all available packages' names. """
        if name is None:
            name = self.module.__name__
        return [pkg for pkg in self.get_all_packages() if name.lower() in self.get_dependencies(name=pkg)]


























